# -*- perl -*-

use strict;
use warnings;
#use FindBin '$Bin';
#use lib "$Bin/lib";

use Test::More;
eval "use Test::Pod 1.00";
plan skip_all => "Test::Pod 1.00 required for testing POD" if $@;
all_pod_files_ok();

done_testing;